from typing import Callable, Union


class StoredFunction:
    """Class that is used to store a function, its arguments, and a name."""

    def __init__(
        self, function: Callable, args: Union[list, dict], kwargs: dict, name: str
    ):
        """inits a StoredFunction Class.

        Args:
           args: A list that will contain positional arguments.
               In case the function only contains keyword arguments,
               This will be dictionary.

           kwargs: a dictionary that will contain keyword arguments
               that will be passed to the function.

           name: a descriptive name for the function.
               defaults to the name of the function.
        """

        self.function = function
        self.args = self._get_args(args)
        self.kwargs = self._get_kwargs(args, kwargs)
        self.name = self._create_name(name)

    def _get_args(self, args):
        if args and not isinstance(args, dict):
            return args
        else:
            return []

    def _get_kwargs(self, args, kwargs):
        # In case only keyword arguments are passed,
        # they are likely to be passed to args.
        if args and isinstance(args, dict):
            return args
        elif kwargs and isinstance(kwargs, dict):
            return kwargs
        else:
            return {}

    def _create_name(self, name: str):
        """If name is not specified by the user,
        then the name defaults to the name of the function.
        For e.g.
        def my_func():
            l = [i for i in range(100)]

        Defaults to my_func
        """
        if not name:
            return self.function.__name__
        else:
            return name



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ptimeit']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'pretty-timeit',
    'version': '0.1.0',
    'description': 'timeit, but with the interface it should have had.',
    'long_description': '# ptimeit\n[![codecov](https://codecov.io/gh/AbhinavOmprakash/pretty-timeit/branch/main/graph/badge.svg?token=QCRpIcv84o)](https://codecov.io/gh/AbhinavOmprakash/pretty-timeit)\n\n***timeit, but with the interface it should have had.***   \n\n## Why did I write this?\ntimeit is a great module, but whenever I wanted to use it, I had to look up the syntax, and even after that it was tricky to get things working. I wanted something that had the similar functionality of timeit but an easier, more intuitive syntax.   \nUnder the hood, I take the same approach as timeit does-\n- Everything is imported once.\n- The garbage collector is disabled when the function is run.\n- By default the function runs 1 milllion times.\n\n## Installation\n\nWith pip.\n\n```Bash\n$ pip install pretty-timeit\n```\n\nWith poetry.\n\n```Bash\n$ poetry add pretty-timeit\n```\n\n## Usage.\n\n### A simple example.\n\n```Python\nfrom ptimeit import timethis, Timer\n\n# although this is a decorator, your original function will not be modified.\n@timethis() \ndef function_to_be_timed():\n    lst = [i for i in range(10)]\n\nTimer.run() # Call this at the end of the file.\n```\n\noutput\n```Console\nname                 | Execution time\nfunction_to_be_timed | 0.5608107\n```\nNote: By default `Timer.run()` Prints the results to the console. \nIf you want it to return a string, do this.\n\n```Python\nTimer.run(print_results=False) # returns a string\n```\n\n\n### Mixing positional arguments with keyword arguments.\nPositional Arguments must be passed inside a list even if there is one argument.  \nKeyword arguments must be passed inside a dictionary.   \nNote the order follows the common idiom of `function(*args, **kwargs)`.\n\n```Python\n@timethis( [10], {"second_count_up_to":10} )\ndef function_to_be_timed(count_up_to, second_count_up_to=100):\n    lst = [i for i in range(count_up_to)]\n    lst2 = [i for i in range(second_count_up_to)]\n\nTimer.run()\n```\n\n### Adding custom descriptive names to your functions\nIf you want to see a different name other than the function name in the final report you can pass that to the decorator using `name=""` argument.\n```Python\n@timethis([10], name="A list comprehension that counts up to 10")\ndef function_to_be_timed(count_up_to):\n    lst = [i for i in range(count_up_to)]\nTimer.run()\n```\n\n```Console\nname                                      | Execution time\nA list comprehension that counts up to 10 | 0.5393135\n``` \n\n### Changing the number of times the function is repeated.\nBy default, like timeit, The function is repeated 1 million times,\nbut this can be changed by passing a `repeat=` to `Timer.run()` like this.\n\n```Python\nTimer.run(repeat=100) #The function to be timed will loop for hundred times.\n```\n\n### Comparing multiple functions.\nA common use case that I have found for the timeit module, is to compare\nthe runtime speed of two different algorithms, this is very easy to do in p-timeit.\n\n```Python\n@timethis(100, name="using a for loop")\ndef my_func(count_up_to):\n    lst=[]\n    for i in range( count_up_to): \n        lst.append(i)\n\n@timethis(100, name="using a List comprehension")\ndef my_func_2(count_up_to):\n    lst = [i for i in range(count_up_to)]\n\nTimer.run(compare=True)\n```\nOutput\n```console\nrank | name                       | Execution time\n1    | using a List comprehension | 2.3460704000000003\n2    | using a for loop           | 4.37266\n```\nAs you can see the list comprehension is faster than using a for loop.\n\nNote: you can compare as many functions as you like. Not just two.\nThe `compare=True` flag formats the output, And orders the results from The fastest functions to the slowest.\nYou can still run multiple functions with `Timer.run()`.\n\nHere\'s an example without compare\n\n```python\n@timethis([10], name="using a for loop")\ndef my_func(count_up_to):\n    lst=[]\n    for i in range( count_up_to): \n        lst.append(i)\n\n@timethis([10], name="using a List comprehension")\ndef my_func_2(count_up_to):\n    lst = [i for i in range(count_up_to)]\n\n@timethis([10], name="using two lists")\ndef my_func_3(count_up_to):\n    lst = [i for i in range(count_up_to)]\n    lst2 = [i for i in range(count_up_to)]\n\nTimer.run()\n```\noutput\n```Console\nname                       | Execution time\nusing a for loop           | 0.7021932000000001\nusing a List comprehension | 0.6186622999999999\nusing two lists            | 1.0627897000000002\n```\nNotice that the functions are ordered as they were defined\nand not by execution time.\n',
    'author': 'Abhinav Omprakash',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/AbhinavOmprakash/pretty-timeit',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.6.2,<4.0.0',
}


setup(**setup_kwargs)



import itertools
import gc
import time
from functools import wraps
from typing import Union, Iterable
from .helpers import StoredFunction


class Timer:
    """Class that handles executing the stored functions and timing them.

    Instance Attributes:
        functions_to_be_timed: A list that contains the StoredFunction object.
            The timethis decorator will "collect" functions in this list.

        exec_time: A dictionary that will contain the name of the function,
            and the time it took to execute.
            This will be used for printing out or generating a string
    """

    functions_to_be_timed = []
    exec_time = {}

    @staticmethod
    def run(repeat=1000000, compare=False, print_results=True):
        """calls the _timeit method on stored functions and returns/prints a report.

        Args:
            repeat: The number of times each function will be run.

            compare: If you wish to compare multiple algorithms,
                this flag tells the report to order
                the functions from fastest to slowest.

            print_results: Flag that determines whether the report
                will be returned as a string or be printed."""

        for stored_func in Timer.functions_to_be_timed:
            time = Timer._timeit(repeat, stored_func)
            Timer.exec_time[stored_func.name] = str(time)

        if print_results:
            print(Timer._format_report(compare))
        else:
            return Timer._format_report(compare)

    @staticmethod
    def _timeit(repeat, stored_func):
        func = stored_func.function
        args = stored_func.args
        kwargs = stored_func.kwargs

        gcold = gc.isenabled()
        gc.disable()

        try:
            start = time.perf_counter()
            for _ in itertools.repeat(None, repeat):
                func(*args, **kwargs)
            end = time.perf_counter()

        finally:
            if gcold:
                gc.enable()

        return end - start

    @staticmethod
    def _format_report(compare=False) -> str:
        def normal_report():
            report = f"name{(space_for_name-4)*space_char} | Execution time\n"  # 4 is the lenght of name
            for name, time in Timer.exec_time.items():
                report += f"{name}{(space_for_name-len(name))*space_char} | {time}\n"
            return report

        def ranked_report():
            report = f"rank | name{(space_for_name-4)*space_char} | Execution time\n"  # 4 is the lenght of name
            rank = 1
            for name, time in sorted(
                Timer.exec_time.items(), key=lambda items: items[1]
            ):  # sort by fastest executing function
                report += f"{rank}{(4-len(str(rank)))*space_char} | {name}{(space_for_name-len(name))*space_char} | {time}\n"
                rank += 1
            return report

        space_for_name = max(
            map(len, Timer.exec_time.keys())
        )  # find length of longest function name
        space_char = " "
        if compare:
            return ranked_report()
        else:
            return normal_report()


def timethis(args: Union[list, dict] = None, kwargs: dict = None, name: str = None):
    """Decorator that will stores the function in Timer.functions_to_be_timed.

    Args:
        args: A list that will contain positional arguments.
            In case the function only contains keyword arguments,
            This lets the user pass in a dictionary without passing in an empty list.
            So the user doesn't have to do this @timethis([], {"keyword":" argument"})
            And instead can do this @timethis({"keyword":" argument"}).

        kwargs: a dictionary that will contain keyword arguments
            that will be passed to the function.

        name: a descriptive name for the function.

    """
    if args and not isinstance(args, list) and not isinstance(args, dict):
        raise TypeError(
            f"Expected a List or Dict but got {type(args).__name__}. Did you forget To put the arguments in a list?"
        )

    if kwargs and not isinstance(kwargs, dict):
        if isinstance(kwargs, str):
            raise TypeError(
                f"Expected a Dictionary but got {type(args).__name__}. Did mean to use name={kwargs}?"
            )
        else:
            raise TypeError(f"Expected a Dictionary but got {type(args).__name__}.")

    def inner_decorator(func):
        f = StoredFunction(func, args, kwargs, name)
        Timer.functions_to_be_timed.append(f)

        @wraps(func)
        def wrapped(*args, **kwargs):
            return func(*args, **kwargs)

        return wrapped

    return inner_decorator
